class FramesExtractorOutPutDirDoesNotExist(Exception):
    """The frames output directory passed to the frame extractor does not exist."""


class FFmpegError(Exception):
    """Base error for the FFmpeg software."""


class FFmpegNotFound(FFmpegError):
    """FFmpeg is either not installed or not in the executable path of the system."""


class FFmpegFailedToExtractFrames(FFmpegError):
    """FFmpeg failed to extract any frame at all. Maybe the input video is damaged or corrupt."""


class DownloadOutPutDirDoesNotExist(Exception):
    """Download output directory does not exist"""


class DownloadFailed(Exception):
    """Download failed"""



import os
import re
import shlex
from shutil import which
from subprocess import PIPE, Popen, check_output
from typing import Optional, Union

from .exceptions import (
    FFmpegError,
    FFmpegFailedToExtractFrames,
    FFmpegNotFound,
    FramesExtractorOutPutDirDoesNotExist,
)
from .utils import does_path_exists

# python module to extract the frames from the input video.
# Uses the FFmpeg Software to extract the frames.


class FramesExtractor:

    """
    Extract frames from the input video file and save at the output directory(frame storage directory).
    """

    def __init__(
        self,
        video_path: str,
        output_dir: str,
        interval: Union[int, float] = 1,
        ffmpeg_path: Optional[str] = None,
    ) -> None:
        """
        Raises Exeception if video_path does not exists.
        Raises Exeception if output_dir does not exists or if not a directory.

        Checks  the ffmpeg installation and the path; thus ensure that we can use it.

        :return: None

        :rtype: NoneType

        :param video_path: absolute path of the video

        :param output_dir: absolute path of the directory
                           where to save the frames.

        :param interval: interval is seconds. interval must be an integer.
                         Extract one frame every given number of seconds.
                         Default is 1, that is one frame every second.

        :param ffmpeg_path: path of the ffmpeg software if not in path.

        """
        self.video_path = video_path
        self.output_dir = output_dir
        self.interval = interval
        self.ffmpeg_path = ""
        if ffmpeg_path:
            self.ffmpeg_path = ffmpeg_path

        if not does_path_exists(self.video_path):
            raise FileNotFoundError(
                f"No video found at '{self.video_path}' for frame extraction."
            )

        if not does_path_exists(self.output_dir):
            raise FramesExtractorOutPutDirDoesNotExist(
                f"No directory called '{self.output_dir}' found for storing the frames."
            )

        self._check_ffmpeg()

        self.extract()

    def _check_ffmpeg(self) -> None:
        """
        Checks the ffmpeg path and runs 'ffmpeg -version' to verify that the
        software, ffmpeg is found and works.

        :return: None

        :rtype: NoneType
        """

        if not self.ffmpeg_path:

            if not which("ffmpeg"):

                raise FFmpegNotFound(
                    "FFmpeg is not on the system path. Install FFmpeg and add it to the path."
                    + "Or you can also pass the path via the 'ffmpeg_path' parameter."
                )
            else:

                self.ffmpeg_path = str(which("ffmpeg"))

        # Check the ffmpeg
        try:
            # check_output will raise FileNotFoundError if it does not finds the ffmpeg
            output = check_output([str(self.ffmpeg_path), "-version"]).decode()

        except FileNotFoundError:
            raise FFmpegNotFound(f"FFmpeg not found at '{self.ffmpeg_path}'.")

        else:

            if "ffmpeg version" not in output:
                raise FFmpegError(
                    f"ffmpeg at '{self.ffmpeg_path}' is not really ffmpeg. Output of ffmpeg -version is \n'{output}'."
                )

    @staticmethod
    def detect_crop(
        video_path: Optional[str] = None,
        frames: int = 3,
        ffmpeg_path: Optional[str] = None,
    ) -> str:
        """
        Detects the the amount of cropping to remove black bars.

        The method uses [ffmpeg.git] / libavfilter /vf_cropdetect.c
        to detect_crop for some fixed intervals.

        The mode of the detected crops is selected as the crop required.

        :return: FFmpeg argument -vf filter and confromable crop parameter.

        :rtype: str
        """

        # We look upto the 120th minute into the video to detect the most
        # precise crop value
        time_start_list = [
            2,
            5,
            10,
            20,
            40,
            100,
            300,
            600,
            1200,
            2400,
            7200,
            14400,
        ]

        crop_list = []

        for start_time in time_start_list:

            command = f'"{ffmpeg_path}" -ss {start_time} -i "{video_path}" -vframes {frames} -vf cropdetect -f null -'

            process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)

            output, error = process.communicate()

            matches = re.findall(
                r"crop\=[0-9]{1,4}:[0-9]{1,4}:[0-9]{1,4}:[0-9]{1,4}",
                (output.decode() + error.decode()),
            )

            for match in matches:
                crop_list.append(match)

        mode = None
        if len(crop_list) > 0:
            mode = max(crop_list, key=crop_list.count)

        crop = " "
        if mode:
            crop = f" -vf {mode} "

        return crop

    def extract(self) -> None:
        """
        Extract the frames at every n seconds where n is the
        integer set to self.interval.

        :return: None

        :rtype: NoneType
        """

        ffmpeg_path = self.ffmpeg_path
        video_path = self.video_path
        output_dir = self.output_dir

        if os.name == "posix":
            ffmpeg_path = shlex.quote(self.ffmpeg_path)
            video_path = shlex.quote(self.video_path)
            output_dir = shlex.quote(self.output_dir)

        crop = FramesExtractor.detect_crop(
            video_path=video_path, frames=3, ffmpeg_path=ffmpeg_path
        )

        command = (
            f'"{ffmpeg_path}"'
            + " -i "
            + f'"{video_path}"'
            + f"{crop}"
            + " -s 64x64 "
            + " -r "
            + str(self.interval)
            + " "
            + '"'
            + output_dir
            + "video_frame_%07d.jpeg"
            + '"'
        )

        process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
        output, error = process.communicate()

        ffmpeg_output = output.decode()
        ffmpeg_error = error.decode()

        if len(os.listdir(self.output_dir)) == 0:

            raise FFmpegFailedToExtractFrames(
                f"FFmpeg could not extract any frames.\n{command}\n{ffmpeg_output}\n{ffmpeg_error}"
            )


def main():
    video_path = (
        "/home/akamhy/projects/videofingerprint/sample_videos/concatenated.webm"
    )
    output_dir = "/home/akamhy/projects/videofingerprint/frames/concatenated_frames/"
    fe = FramesExtractor(video_path=video_path, output_dir=output_dir, interval=1)


if __name__ == "__main__":
    main()



__title__ = "videofingerprint"
__description__ = (
    "Relative Luminance gradient and Dominant Color of Frames (RLGDCF) "
    + "based Digital Video Fingerprinting. "
)

__url__ = "https://akamhy.github.io/videofingerprint/"
__version__ = "0.0.1"
__status__ = "production"
__author__ = "Akash Mahanty"
__author_email__ = "akamhy@yahoo.com"
__license__ = "MIT"
__copyright__ = "Copyright 2022 Akash Mahanty"



"""
Relative Luminance gradient and Dominant Color of Frames (RLGDCF)
based Digital Video Fingerprinting
"""

from .__version__ import (
    __author__,
    __author_email__,
    __copyright__,
    __description__,
    __license__,
    __status__,
    __title__,
    __url__,
    __version__,
)
from .exceptions import (
    DownloadFailed,
    DownloadOutPutDirDoesNotExist,
    FFmpegError,
    FFmpegFailedToExtractFrames,
    FFmpegNotFound,
    FramesExtractorOutPutDirDoesNotExist,
)
from .rlgdcf_videofingerprint import VideoFingerprint
from .videoduration import video_duration



import os
import random
import re
import shutil
from pathlib import Path
from typing import List

from imagedominantcolor import DominantColor
from PIL import Image

from .downloader import Download
from .framesextractor import FramesExtractor
from .utils import (
    create_and_return_temporary_directory,
    does_path_exists,
    get_list_of_all_files_in_dir,
)
from .videoduration import video_duration as videoduration


class VideoFingerprint:
    """
    Relative Luminance gradient and Dominant Color of Frames (RLGDCF)
    based Digital Video Fingerprinting
    """

    storage_path = None
    download_worst = False
    dominant_color = DominantColor
    dominant_color.resize_value = 16
    dominant_color.minimum_percent_difference_of_rgb = 10
    previous_color = 0

    def __init__(self, path=None, url=None):
        self.path = path
        self.url = url
        self.fingerprint = ""
        self.luminance_fingerprint = ""
        self.chrominance_fingerprint = ""
        self.video_path: str = ""
        self.task_uid = VideoFingerprint.get_task_uid()
        self.create_required_dirs_and_check_for_errors()
        self.copy_video_to_video_dir()
        FramesExtractor(self.video_path, self.frames_dir)
        self.frames = get_list_of_all_files_in_dir(self.frames_dir)
        self.generate_fingerprint()

    @property
    def video_duration():
        return videoduration(self.video_path)

    def copy_video_to_video_dir(self):

        if self.path:
            # create a copy of the video at self.storage_path
            match = re.search(r"\.([^.]+$)", self.path)

            if match:
                extension = match.group(1)

            else:
                raise ValueError("File name (path) does not have an extension.")

            self.video_path = os.path.join(self.video_dir, (f"video.{extension}"))

            shutil.copyfile(self.path, self.video_path)

        if self.url:

            Download(
                self.url,
                self.video_download_dir,
                worst=self.download_worst,
            )

            downloaded_file = get_list_of_all_files_in_dir(self.video_download_dir)[0]
            match = re.search(r"\.(.*?)$", downloaded_file)

            extension = "mkv"

            if match:
                extension = match.group(1)

            self.video_path = f"{self.video_dir}video.{extension}"

            shutil.copyfile(downloaded_file, self.video_path)

    def create_required_dirs_and_check_for_errors(self) -> None:
        if not self.path and not self.url:
            raise DidNotSupplyPathOrUrl(
                "You must specify either a path or an URL of the video."
            )

        if self.path and self.url:
            raise ValueError("Specify either a path or an URL and NOT both.")

        if not VideoFingerprint.storage_path:
            VideoFingerprint.storage_path = create_and_return_temporary_directory()
        if not does_path_exists(VideoFingerprint.storage_path):
            raise StoragePathDoesNotExist(
                f"Storage path '{VideoFingerprint.storage_path}' does not exist."
            )

        os_path_sep = os.path.sep

        VideoFingerprint.storage_path = os.path.join(
            VideoFingerprint.storage_path, (f"{self.task_uid}{os_path_sep}")
        )

        self.video_dir = os.path.join(
            VideoFingerprint.storage_path, (f"video{os_path_sep}")
        )
        Path(self.video_dir).mkdir(parents=True, exist_ok=True)

        self.video_download_dir = os.path.join(
            VideoFingerprint.storage_path, (f"downloadedvideo{os_path_sep}")
        )
        Path(self.video_download_dir).mkdir(parents=True, exist_ok=True)

        self.frames_dir = os.path.join(
            VideoFingerprint.storage_path, (f"frames{os_path_sep}")
        )
        Path(self.frames_dir).mkdir(parents=True, exist_ok=True)

    def image_luminance(self, image_path, width=1, height=1):
        return list(
            Image.open(image_path)
            .convert("L")
            .resize((width, height), Image.ANTIALIAS)
            .getdata()
        )[0]

    def color_case(self, image_path):
        current_color = self.image_luminance(image_path)
        if VideoFingerprint.previous_color <= current_color:
            case = 0
        else:
            case = 1
        VideoFingerprint.previous_color = current_color
        return case

    def generate_fingerprint(self):

        for frame in self.frames:
            dc = VideoFingerprint.dominant_color(image_path=frame).dominant_color
            self.chrominance_fingerprint += dc
            case = self.color_case(frame)
            self.luminance_fingerprint += str(case)
            case_dc = dc.upper() if case == 1 else dc
            self.fingerprint += case_dc
        return self

    def get_task_uid() -> str:
        sys_random = random.SystemRandom()

        return "".join(
            sys_random.choice(
                "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
            )
            for _ in range(20)
        )
